import path from 'node:path';
import {
    Cookie,
    CookieJar
} from "tough-cookie";
import {
    FileCookieStore
} from "tough-cookie-file-store";
import {
    isBefore
} from 'date-fns'
import { Logger } from 'winston'
import { IAppContext } from './appContext'
import { chaturbateRootUrl } from './constants.js'

export interface IPushServiceAuth {
    token: string;
    channels: Record<string, string>;
    failures: Record<string, any>;
    token_request: {
        keyName: string;
        clientId: string;
        ttl: number;
        nonce: string;
        capability: string;
        timestamp: number;
        mac: string;
    };
    client_id: string;
    settings: {
        backend: string;
        flags: {
            pm_enabled: boolean;
            wowza_disabled: boolean;
            userlist_enabled: boolean;
            verify_enabled: boolean;
            fallback_eligible: boolean;
            is_live: boolean;
        };
        rest_host: string;
        realtime_host: string;
        fallback_hosts: string[];
    };
}


export default class ChaturbateAuth {
    private appContext: IAppContext;
    private logger: Logger;
    public tokenRequest: any;
    private cookieJar: CookieJar;
    private pushServiceAuth: IPushServiceAuth;

    constructor(appContext) {
        this.appContext = appContext;
        this.logger = appContext.logger;
        this.tokenRequest;
        this.pushServiceAuth;
        this.cookieJar = new CookieJar(new FileCookieStore(path.join(appContext.dataDir, "cookie.json")));
    }



    async fetchCookiesIfNeeded() {
        let cookies = await this.cookieJar.getCookies(chaturbateRootUrl);
        let tokenCookie = ChaturbateAuth.getTokenCookie(cookies)
        this.logger.log({
            level: 'debug',
            message: `tokenCookie is as follows. ${tokenCookie}`
        })
        if (
            tokenCookie === undefined ||
            isBefore(tokenCookie.expires, new Date())
        ) {
            this.logger.log({
                level: 'debug',
                message: 'tokenCookie is either undefined or expired. Getting new cookie.'
            })
            let res = await this.appContext.gotClient(chaturbateRootUrl, {
                "headers": {
                    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8",
                    "Accept-Language": "en-US,en;q=0.5",
                    "Upgrade-Insecure-Requests": "1",
                    "Sec-Fetch-Dest": "document",
                    "Sec-Fetch-Mode": "navigate",
                    "Sec-Fetch-Site": "none",
                    "Sec-Fetch-User": "?1"
                },
                "method": "GET",
            })

            const rawHeaders = res.headers
            if (Array.isArray(rawHeaders["set-cookie"]))
                cookies = rawHeaders["set-cookie"].map(Cookie.parse);
            else cookies = [Cookie.parse(rawHeaders["set-cookie"])];

            for (const cookie of cookies) {
                this.logger.log({ level: 'debug', message: `adding cookie ${JSON.stringify(cookie)} to the cookieJar` })
                await this.cookieJar.setCookie(cookie, chaturbateRootUrl)
            }
        } else {
            this.logger.log({ level: 'debug', message: `🍪✅ cookies are valid.` })
        }
    }

    static getPresenceId(): string {
        return Math.random().toString(36).substring(2)
    }

    // the number suffixes are assumed to be session ids.
    // opening multiple tabs of the same room will not increment the :n suffix.
    // however, opening different rooms in multiple tags will increment the :n suffix
    // maybe Ably does request deduplication for same sesionIds???
    // we don't need to compute the sessionId, as pushServiceAuth response contains it.
    static getPermissionsForm(roomIds: string[], topics: string[], csrfToken: string): FormData {
        if (!roomIds || roomIds.length === 0) {
            throw new Error('roomIds is required but it was undefined or empty');
        }

        const generateTopic = (topicKey: string, topicValue: string) => {
            return JSON.stringify({ [topicKey]: topicValue });
        };

        let form = new FormData();
        let formTopics = {};

        roomIds.forEach((roomId) => {

            if (topics.includes('message')) {
                formTopics[`RoomMessageTopic#RoomMessageTopic:${roomId}`] = { "broadcaster_uid": roomId }
            }

            // @todo this one is somehow bad https://gitea.futureporn.net/futureporn/futureporn-scout/issues/3
            // if (topics.includes('silence')) {
            //     formTopics[`RoomSilenceTopic#RoomSilenceTopic:${roomId}`] = { "broadcaster_uid": roomId }
            // }

            if (topics.includes('status')) {
                formTopics[`RoomStatusTopic#RoomStatusTopic:${roomId}`] = { "broadcaster_uid": roomId }
            }

            // @todo this one is somehow bad https://gitea.futureporn.net/futureporn/futureporn-scout/issues/3
            // if (topics.includes('title')) {
            //     formTopics[`RoomTitleChangeTopic#RoomTitleChangeTopic:${roomId}`] = { "broadcaster_uid": roomId }
            // }

            if (topics.includes('tip')) {
                formTopics[`RoomTipAlertTopic#RoomTipAlertTopic:${roomId}`] = { "broacaster_uid": roomId }
            }

            if (topics.includes('password')) {
                formTopics[`RoomPasswordProtectedTopic#RoomPasswordProtectedTopic:${roomId}`] = { "broadcaster_uid": roomId }
            }

        });

        form.append('topics', JSON.stringify(formTopics))

        form.append('csrfmiddlewaretoken', csrfToken);

        return form;
    }
    
    static getTokenCookie(cookies) {
        return cookies.find((c) => c.key == 'csrftoken');
    }

    static getSbrCookie(cookies) {
        return cookies.find((c) => c.key == 'sbr');
    }

    async getTokenRequest(url: string, form: FormData, cookie: string) {
        if (!url) throw new Error('url must be passed to getTokenRequest');
        if (!form) throw new Error('FormData must be passed to getTokenRequest');
        if (!cookie) throw new Error('cookie must be passed to getTokenRequest');
        this.pushServiceAuth = await this.getPushServiceAuth(url, form, cookie)
        return this.pushServiceAuth.token_request
    }

    async getCsrfToken(roomUrl) {
        let cookies = await this.cookieJar.getCookies(roomUrl);
        let tokenCookie = ChaturbateAuth.getTokenCookie(cookies)
        return tokenCookie.value
    }

    /**
     * make a request against CB's server which
     *   - verifies custom credentials
     *   - issues signed Ably TokenRequest
     * 
     * The response contains the following
     *   - token
     *   - channels
     *   - failures
     *   - token_request <-- signed Ably TokenRequest
     *   - client_id
     *   - settings
     *
     * GET https://chaturbate.com/push_service/auth/
     *  
     * more info-- https://ably.com/docs/core-features/authentication#token-authentication
     */
    async getPushServiceAuth(
        roomUrl: string,
        form: FormData,
        cookieString: string
    ) {


        // https://faqs.ably.com/40104-timestamp-not-current
        // this was caused by signed token request re-use after it had expired.
        // we must not cache the signed token request


        // https://help.ably.io/error/40102
        // caused by mismatching client_id, when using a new client_id issued by CB pushServiceAuth with an old ably connection
        // (possibly) caused by not sending Cookie header during CB pushServiceAuth
        //            thus leading to CB thinking we're a new anon and issuing a new client_id
        //            @todo please verify if this is the cause

        // 09:44:42.182 Ably: Transport.onIdleTimerExpire(): No activity seen from realtime in 25113ms; assuming connection has dropped
        // saw this when my internet died. unsure of whether or not the connection comes back online.
        // if it doesn't come back online, this could bite me in the future.
        // this is a thing we can test by airgapping.
        // if there's a way to implement an onIdleTimerExpire callback, that would be the solution.
        // (solved by https://github.com/insanity54/futureporn/issues/208)

        this.logger.log({
            level: 'debug',
            message: `getting pushServiceAuth.`
        })

        this.logger.log({
            level: 'debug',
            message: `using cookie string ${cookieString}`
        })


        this.pushServiceAuth = await this.appContext.gotClient("https://chaturbate.com/push_service/auth/", {
            "headers": {
                "Cookie": cookieString,
                "Accept": "*/*",
                "Accept-Language": "en-US,en;q=0.5",
                "X-Requested-With": "XMLHttpRequest",
                "Sec-Fetch-Dest": "empty",
                "Sec-Fetch-Mode": "cors",
                "Sec-Fetch-Site": "same-origin",
                "Sec-GPC": "1",
                "Referrer": roomUrl,
            },
            "body": form,
            "method": "POST"
        }).json()

        this.logger.log({
            level: 'debug',
            message: JSON.stringify(this.pushServiceAuth)
        })
        return this.pushServiceAuth
    }

}