
import {
  FormData
} from 'formdata-polyfill/esm.min.js'
import cheerio from 'cheerio'
import * as Ably from 'ably/promises.js';
import ChaturbateAuth from './ChaturbateAuth.js'
import { Logger } from 'winston'
import { IAppContext } from './appContext.js';
import { IChannelMap, IPushServiceAuth, getPushServiceAuth } from './headless.js';
import { getSuperRealtimeClient } from './realtime.js';
import { z } from 'zod';

export interface IRoomOptions {
  chaturbateAuth?: any;
  name: string;
  onStatus?: any;
  onMessage?: any;
  onSilence?: any;
  onTitle?: any;
  onTip?: any;
  onPassword?: any;
  onNotice?: any;
  id?: string;
}

export enum PubOrPass {
  Public = 'PUBLIC',
  Password = 'PASSWORD'
}

export interface ChannelMap {
  [topic: string]: string;
}

export interface ChannelsRequest {
  [topic: string]: {
    broadcaster_uid?: string;
    user_uid?: string;
  };
}

interface RoomTopic {
  broadcaster_uid?: string;
  user_uid?: string;
}

interface Topics {
  [topicId: string]: RoomTopic;
}

// export interface IRoomMessage {
//   name: string;
//   id: string;
//   encoding: null;
//   data: {
//     tid: string;
//     ts: number;
//     _topic: string;
//     message: string;
//     font_family: string;
//     font_color: string;
//     background: string;
//     id: string;
//     from_user: {
//       username: string;
//       gender: string;
//       is_broadcaster: boolean;
//       in_fanclub: boolean;
//       is_following: boolean;
//       is_mod: boolean;
//       has_tokens: boolean;
//       tipped_recently: boolean;
//       tipped_alot_recently: boolean;
//       tipped_tons_recently: boolean;
//     };
//     status?: string;
//     method: string;
//     pub_ts: number;
//   };
// }



export default class Room {
  public id: string | Promise<string>;
  private logger: Logger;
  private appContext: IAppContext;
  private chaturbateAuth: ChaturbateAuth;
  public name: string;
  public url: string;
  private realtimeClient: Ably.Realtime | null;
  private csrfToken: string | null;
  private tokenRequest: any;
  private ablyTokenRequest: any;
  private realtimeHost: string | null;
  private fallbackHosts: string[] | null;
  private pushServiceAuth: any;
  private onStatus: ((message: Ably.Types.Message) => void);
  private onMessage: ((message: Ably.Types.Message) => void);
  private onSilence: ((message: Ably.Types.Message) => void);
  private onTitle: ((message: Ably.Types.Message) => void);
  private onTip: ((message: Ably.Types.Message) => void);
  private onPassword: ((message: Ably.Types.Message) => void);
  private onNotice: ((message: Ably.Types.Message) => void);
  private dossier: any;
  public errors: string[];
  public pubOrPass: PubOrPass;
  public channelsRequest: ChannelsRequest;
  public channelMap: ChannelMap;


  constructor(appContext: IAppContext, opts: IRoomOptions) {
    if (opts?.name === undefined) throw new Error('Room constructor requires a room name.')
    Room.roomNameSchema.parse(opts.name);
    this.logger = appContext.logger // this needs to be defined before this.id
    this.appContext = appContext
    this.chaturbateAuth = opts.chaturbateAuth;
    this.name = opts.name
    this.url = 'https://chaturbate.com/' + this.name + '/'
    this.csrfToken = null
    this.tokenRequest = null
    this.ablyTokenRequest = null
    this.realtimeHost = null
    this.realtimeClient = null
    this.fallbackHosts = null
    this.pushServiceAuth = null
    this.onStatus = opts.onStatus;
    this.onMessage = opts.onMessage;
    this.onSilence = opts.onSilence;
    this.onTitle = opts.onTitle;
    this.onTip = opts.onTip;
    this.onPassword = opts.onPassword;
    this.onNotice = opts.onNotice;
    this.dossier = null
    this.errors = [];
    this.pubOrPass = PubOrPass.Public
    this.channelsRequest;
    this.channelMap;


    // We put this last because they depend on other opts
    // before making a network request
    this.id = opts?.id || this.getRoomId()
  }


  handleError(err: string) {
    this.logger.log({ level: 'debug', message: `handling room error. ${err}` })
    this.errors.push(err)
    if (/room has a password/.test(err)) this.pubOrPass = PubOrPass.Password
  }

  // depends on initialRoomDossier existing, which is missing when room is passworded
  // will return null if dossier is missing.
  async getRoomId(): Promise<string | null> {
    try {
      const dossier = await this.getInitialRoomDossier()
      return this.id = dossier.room_uid
    } catch (e) {
      this.logger.log({ level: 'warn', message: `error while getting room ID. ${e}` })
      return this.id = null
    }
  }

  async getLocalPushServiceAuth(): Promise<IPushServiceAuth> {
    if (!!this.pushServiceAuth) return this.pushServiceAuth;
    try {
      const psa = await getPushServiceAuth(this.url);
      this.channelMap = psa.channels;
      return this.pushServiceAuth = psa;
    } catch (e) {
      this.logger.log({ level: 'warn', message: `error while getting psa. ${e}` })
    }
  }


  async getInitialRoomDossier(): Promise<any | null> {
    try {
      const res = await this.appContext.gotClient(this.url);
      const $ = cheerio.load(res.body);
      let rawScript = $('script:contains(window.initialRoomDossier)').html();
      if (!rawScript) {
        throw new Error('window.initialRoomDossier is null. This could mean the channel is in password mode');
      }
      let rawDossier = rawScript.slice(rawScript.indexOf('"'), rawScript.lastIndexOf('"') + 1);
      let dossier = JSON.parse(JSON.parse(rawDossier));
      return dossier;
    } catch (error) {
      // Handle the error gracefully
      this.logger.log({ level: 'warn', message: `Error fetching initial room dossier: ${error.message}` });
      return null; // Or any other appropriate action you want to take
    }
  }

  static roomNameSchema = z.string().regex(/^[a-zA-Z0-9_]+$/);


  static getIdFromTopic(topic: string): string {
    const parts = topic.split(':');
    if (/^\d+$/.test(parts.at(-1))) {
      return parts.at(-2)
    } else {
      return parts.at(-1)
    }
  }


  // /**
  //  * 
  //  * The list of default topics which are always subscribed to
  //  * 
  //  */
  // public async getDefaultTopics(): Promise<string[]> {
  //   await this.id
  //   if (!this.id) throw new Error(`cannot get Room's default topics because a room.id could not be fetched`);
  //   return [`RoomStatusTopic#RoomStatusTopic:${this.id}`]
  // }

  // public async getAvailableTopics(): Promise<string[]> {
  //   await this.id
  //   if (!this.id) throw new Error(`cannot get Room's available topics because room.id could not be fetched`)
  //   const defaultTopics = await this.getDefaultTopics()
  //   return defaultTopics
  // }


  /**
   * merge several channels requests forms together.
   * useful for combining room permissions forms in order to reduce
   * network requests by Ably realtime client.
   * This lets us use one Ably realtime client with permissions for
   * all the rooms we are interested in. 
   */
  static mergeChannelsRequests(requests: ChannelsRequest[]): ChannelsRequest {
    const mergedRequest: ChannelsRequest = {};

    for (const request of requests) {
      for (const topic in request) {
        if (!mergedRequest.hasOwnProperty(topic)) {
          mergedRequest[topic] = {};
        }

        if (request[topic].broadcaster_uid) {
          mergedRequest[topic].broadcaster_uid = request[topic].broadcaster_uid;
        }

        if (request[topic].user_uid) {
          mergedRequest[topic].user_uid = request[topic].user_uid;
        }
      }
    }

    return mergedRequest;
  }



  static formalizeChannelsRequest(channelsRequest: ChannelsRequest, csrfToken: string): FormData {
    let form = new FormData();
    form.append('topics', JSON.stringify(channelsRequest))
    form.append('csrfmiddlewaretoken', csrfToken);
    form.append('backend', 'a')
    // form.append('presence_id', ChaturbateAuth.getPresenceId())
    return form;
  }

  /**
   * filterPermissionsForms
   * 
   * filter permissions forms.
   * accepts a permission form consisting of 
   * permissions for several rooms.
   * returns a single permissions form consisting of only perms
   * belonging to this room ID
   */
  // public filterChannelsRequests (permissionForms: ChannelsRequest): ChannelsRequest {

  // }



  public async getChannelsRequest(): Promise<ChannelsRequest> {
    const roomId: string = await this.id;
    if (!roomId) throw new Error(`cannot get Room's permissions because a room.id could not be fetched`);
    let crForm = {}
    crForm[`RoomMessageTopic#RoomMessageTopic:${roomId}`] = { "broadcaster_uid": roomId }
    // crForm[`RoomSilenceTopic#RoomSilenceTopic:${roomId}`] = { "broadcaster_uid": roomId }
    crForm[`RoomStatusTopic#RoomStatusTopic:${roomId}`] = { "broadcaster_uid": roomId }
    // crForm[`RoomTitleChangeTopic#RoomTitleChangeTopic:${roomId}`] = { "broadcaster_uid": roomId }
    crForm[`RoomTipAlertTopic#RoomTipAlertTopic:${roomId}`] = { "broadcaster_uid": roomId }
    crForm[`RoomPasswordProtectedTopic#RoomPasswordProtectedTopic:${roomId}`] = { "broadcaster_uid": roomId }
    this.channelsRequest = crForm
    return this.channelsRequest
  }


  // public getChannelsRequest(topics: string[]): ChannelsRequest {
  //   let crForm = {}
  //   if (topics.includes('message')) {
  //     crForm[`RoomMessageTopic#RoomMessageTopic:${this.id}`] = { "broadcaster_uid": this.id }
  //   }

  //   if (topics.includes('silence')) {
  //     crForm[`RoomSilenceTopic#RoomSilenceTopic:${this.id}`] = { "broadcaster_uid":this.id }
  //   }

  //   if (topics.includes('status')) {
  //     crForm[`RoomStatusTopic#RoomStatusTopic:${this.id}`] = { "broadcaster_uid":this.id }
  //   }

  //   if (topics.includes('title')) {
  //     crForm[`RoomTitleChangeTopic#RoomTitleChangeTopic:${this.id}`] = { "broadcaster_uid": this.id }
  //   }

  //   if (topics.includes('tip')) {
  //     crForm[`RoomTipAlertTopic#RoomTipAlertTopic:${this.id}`] = { "broacaster_uid": this.id }
  //   }

  //   if (topics.includes('password')) {
  //     crForm[`RoomPasswordProtectedTopic#RoomPasswordProtectedTopic:${this.id}`] = { "broadcaster_uid": this.id}
  //   }
  //   return this.channelsRequest = crForm
  // }

  // the number suffixes are assumed to be session ids.
  // opening multiple tabs of the same room will not increment the :n suffix.
  // however, opening different rooms in multiple tags will increment the :n suffix
  // maybe Ably does request deduplication for same sesionIds???
  // we don't need to compute the sessionId, as pushServiceAuth response contains it.
  static getPermissionsForm(roomIds: string[], topics: string[], csrfToken: string): FormData {
    if (!roomIds || roomIds.length === 0) {
      throw new Error('roomIds is required but it was undefined or empty');
    }

    const generateTopic = (topicKey: string, topicValue: string) => {
      return JSON.stringify({ [topicKey]: topicValue });
    };

    let form = new FormData();
    let formTopics = {};

    roomIds.forEach((roomId) => {

      if (topics.includes('message')) {
        formTopics[`RoomMessageTopic#RoomMessageTopic:${roomId}`] = { "broadcaster_uid": roomId }
      }

      if (topics.includes('silence')) {
        formTopics[`RoomSilenceTopic#RoomSilenceTopic:${roomId}`] = { "broadcaster_uid": roomId }
      }

      if (topics.includes('status')) {
        formTopics[`RoomStatusTopic#RoomStatusTopic:${roomId}`] = { "broadcaster_uid": roomId }
      }

      if (topics.includes('title')) {
        formTopics[`RoomTitleChangeTopic#RoomTitleChangeTopic:${roomId}`] = { "broadcaster_uid": roomId }
      }

      if (topics.includes('tip')) {
        formTopics[`RoomTipAlertTopic#RoomTipAlertTopic:${roomId}`] = { "broadcaster_uid": roomId }
      }

      if (topics.includes('password')) {
        formTopics[`RoomPasswordProtectedTopic#RoomPasswordProtectedTopic:${roomId}`] = { "broadcaster_uid": roomId }
      }

    });
    form.append('topics', JSON.stringify(formTopics))
    form.append('csrfmiddlewaretoken', csrfToken);
    return form;
  }


  /**
   * Get the ably channel string from memory.
   * we can pass in a generic topic such as 'status'
   * 
   */
  // static getAblyChannelName(channelMap, genericTopic: string): string {
  //   // get the cb topic string
  //   // use the cb topic string to find ably ch name in map
  //   const cbTopic = this.getChaturbateTopicString(genericTopic)
  //   return this.channelMap[cbTopic]
  // }

  static getChannelMapFromPsa(psa: IPushServiceAuth) {
    return psa.channels;
  }

  static getRealtimeChannelNameFromCbChannelName(channelMap: ChannelMap, cbChannelName: string) {
    return channelMap[cbChannelName];
  }




  /**
   * 
   * getChaturbateTopicString
   * 
   * Chaturbate now groups message together. the topics that are grouped are
   * 
   *   * status, message, password
   *   * title, silence
   * 
   * @see https://gitea.futureporn.net/futureporn/futureporn-scout/issues/3
   * 
   * @param genericTopic 
   * @returns 
   */
  getChaturbateTopicString(genericTopic: string): string {
    let ch: string;
    if (genericTopic === 'status' || genericTopic === 'message' || genericTopic === 'password') {
      ch = `RoomStatusTopic#RoomStatusTopic:${this.id}`;
    }
    else if (genericTopic === 'tip') {
      ch = `RoomTipAlertTopic#RoomTipAlertTopic:${this.id}`;
    }
    else if (genericTopic === 'title' || genericTopic === 'silence') {
      ch = `RoomTitleChangeTopic#RoomTitleChangeTopic:${this.id}`;
    }
    return ch;
  }





  async getRealtimeClient(): Promise<Ably.Realtime> {
    if (!!this.realtimeClient) return this.realtimeClient;
    try {
      const psa = await this.getLocalPushServiceAuth();
      const rtc = await getSuperRealtimeClient(this.name, psa);
      return this.realtimeClient = rtc;
    } catch (e) {
      this.logger.log({ level: 'error', message: `failed to get realtimeClient. ${e}` });
    }
  }



  async handleRealtimeEvent(message: Ably.Types.Message): Promise<void> {
    this.appContext.logger.log({ level: 'debug', message: `handling realtime event with _topic=${message?.data?._topic}` });
    const cbTopic = message?.data?._topic;
    if (cbTopic.startsWith('RoomStatusTopic')) {
      this.appContext.logger.log({ level: 'warn', message: `⚠️⚠️⚠️ _topic=${cbTopic} ???? is this needed?` })
      this.onStatus(message);
    } else if (cbTopic.startsWith('RoomMessageTopic')) {
      this.onMessage(message);
    } else if (cbTopic.startsWith('RoomTipAlertTopic')) {
      this.onTip(message);
    } else if (cbTopic.startsWith('RoomTitleChangeTopic')) {
      this.onTitle(message);
    } else if (cbTopic.startsWith('RoomSilenceTopic')) {
      this.onSilence(message);
    } else if (cbTopic.startsWith('RoomPasswordProtectedTopic')) {
      this.onPassword(message);
    } else if (cbTopic.startsWith('RoomNoticeTopic')) {
      this.onNotice(message);
    } else {
      this.appContext.logger.log({ level: 'warn', message: `⚠️⚠️⚠️ thure was an incomming message (${cbTopic}) which we dont know how to handle. ` })
      this.appContext.logger.log({ level: 'warn', message: message })
    }
  }



  async subscribeToRealtimeChannel(realtimeChannelName: string): Promise<void> {
    const realtimeChannel = this.realtimeClient.channels.get(realtimeChannelName);
    this.appContext.logger.log({ level: 'debug', message: `subscribing to ${realtimeChannelName}` });
    realtimeChannel.subscribe((message) => {
      this.appContext.logger.log({ level: 'debug', message: JSON.stringify(message, null, 2) });
      this.handleRealtimeEvent(message);
    });
  }

  deduplicateChannelMap(channelMap: ChannelMap): ChannelMap {
    const uniqueValuesSet = new Set<string>();
    const deduplicatedMap: ChannelMap = {};
  
    for (const topic in channelMap) {
      const value = channelMap[topic];
  
      // Check if the value is not already in the uniqueValuesSet
      if (!uniqueValuesSet.has(value)) {
        // Add the value to the uniqueValuesSet and the deduplicatedMap
        uniqueValuesSet.add(value);
        deduplicatedMap[topic] = value;
      }
    }
  
    return deduplicatedMap;
  }


  /*
    "channels": {
      "RoomTipAlertTopic#RoomTipAlertTopic:VVKHB7L": "room:tip_alert:VVKHB7L",
      "RoomPurchaseTopic#RoomPurchaseTopic:VVKHB7L": "room_unsessioned:grouped:VVKHB7L",
      "RoomFanClubJoinedTopic#RoomFanClubJoinedTopic:VVKHB7L": "room:fanclub:VVKHB7L",
      "RoomMessageTopic#RoomMessageTopic:VVKHB7L": "room:grouped:VVKHB7L:6",
      "GlobalPushServiceBackendChangeTopic#GlobalPushServiceBackendChangeTopic": "global:push_service",
      "RoomAnonPresenceTopic#RoomAnonPresenceTopic:VVKHB7L": "room_anon:presence:VVKHB7L:6",
      "QualityUpdateTopic#QualityUpdateTopic:VVKHB7L": "room_unsessioned:grouped:VVKHB7L",
      "RoomNoticeTopic#RoomNoticeTopic:VVKHB7L": "room_unsessioned:grouped:VVKHB7L",
      "RoomPasswordProtectedTopic#RoomPasswordProtectedTopic:VVKHB7L": "room:grouped:VVKHB7L:6",
      "RoomModeratorPromotedTopic#RoomModeratorPromotedTopic:VVKHB7L": "room_unsessioned:grouped:VVKHB7L",
      "RoomModeratorRevokedTopic#RoomModeratorRevokedTopic:VVKHB7L": "room_unsessioned:grouped:VVKHB7L",
      "RoomStatusTopic#RoomStatusTopic:VVKHB7L": "room:grouped:VVKHB7L:6",
      "RoomTitleChangeTopic#RoomTitleChangeTopic:VVKHB7L": "room_unsessioned:grouped:VVKHB7L",
      "RoomSilenceTopic#RoomSilenceTopic:VVKHB7L": "room_unsessioned:grouped:VVKHB7L",
      "RoomKickTopic#RoomKickTopic:VVKHB7L": "room_unsessioned:grouped:VVKHB7L",
      "RoomUpdateTopic#RoomUpdateTopic:VVKHB7L": "room_unsessioned:grouped:VVKHB7L",
      "RoomSettingsTopic#RoomSettingsTopic:VVKHB7L": "room_unsessioned:grouped:VVKHB7L",
      "RoomEnterLeaveTopic#RoomEnterLeaveTopic:VVKHB7L": "room:enter_leave:VVKHB7L",
      "GameUpdateTopic#GameUpdateTopic:VVKHB7L": "room_unsessioned:grouped:VVKHB7L"
    }
  */
  async subscribeToEvents(): Promise<void> {
    if (!this.channelMap) throw new Error('attempting to subscribeToEvents before we have a channelMap!');
    const channelEntries: [string, string][] = Object.entries(this.channelMap);
    const interestingChannels = channelEntries.filter(([topic]) => {
      return (
        topic.startsWith('RoomStatusTopic') ||
        topic.startsWith('RoomTipAlertTopic') ||
        topic.startsWith('RoomTitleChangeTopic') ||
        topic.startsWith('RoomSilenceTopic') ||
        topic.startsWith('RoomPasswordProtectedTopic')
      )
    }).map((ic) => ic[1]);


    // some of the CB channels map to the same realtime channel name. 
    // Because of this, we could have duplicates. 
    // We need to deduplicate.
    const deduplicatedChannels = [...new Set(interestingChannels)];
    this.appContext.logger.log({ level: 'debug', message: `deduplicatedChannels=${deduplicatedChannels}` })

    for (const channel of deduplicatedChannels) {
      this.appContext.logger.log({ level: 'debug', message: `subscribing to channel=${channel}` });
      this.subscribeToRealtimeChannel(channel);
    };
  }

  /**
   * Begin watching the CB room events.
   */
  async startWatching(): Promise<void> {
    const rtc = await this.getRealtimeClient();
    rtc.connect();
    await rtc.connection.once('connected');
    this.logger.log({ level: 'debug', message: `Connected to ${this.name}` });
    this.subscribeToEvents();
  }

  /**
   * Cease watching the CB room events.
   */
  async stopWatching(): Promise<void> {
    const rtc = await this.getRealtimeClient();
    rtc.close();
    return;
  }








}