import { IAppContext } from "./appContext";
import { isBefore, add, formatISO } from 'date-fns';
import TwitterClient from 'twitter-api-scraper'
import puppeteer from 'puppeteer-extra';
import StealthPlugin from 'puppeteer-extra-plugin-stealth'


export interface ITweet {
    id: string;
    full_text: string;
}

export interface IScraperQuery {
    terms: string;
}

const projektMelodyEpoch = new Date('2020-02-07')

export default class Twitter {

    private appContext: IAppContext;
    private name: string;
    private startDate: Date;

    constructor(appContext, opts) {
        this.appContext = appContext;
        this.name = opts.name;
        this.startDate = opts.startDate || projektMelodyEpoch
    }

    async getAllTweets(): Promise<ITweet[]> {

      puppeteer.use(StealthPlugin())

      // Launch the browser and open a new blank page
      const browser = await puppeteer.launch({ headless: false });
      const page = await browser.newPage();

      // Navigate the page to a URL
      await page.goto(`https://twitter.com/${this.name}`);

      // Set screen size
      await page.setViewport({width: 1366, height: 768});

      // // Type into search box
      // await page.type('.search-box__input', 'automate beyond recorder');

      // // Wait and click on first result
      const searchResultSelector = '.r-1ljd8xs > div:nth-child(1)';
      await page.waitForSelector(searchResultSelector);
      // await page.click(searchResultSelector);

      // // Locate the full title with a unique string
      // const textSelector = await page.waitForSelector(
      //   'text/Customize and automate'
      // );
      // const fullTitle = await textSelector?.evaluate(el => el.textContent);

      // // Print the full title
      // console.log('The title of this blog post is "%s".', fullTitle);
      await new Promise(r => setTimeout(r, 8000));

      await browser.close();

      return []

    }

    async search(
        client: any, 
        scraperQuery: IScraperQuery,
        nextToken: string | null,
        startDate: Date,
        endDate: Date
      ): Promise<ITweet[]> {
        const later = (delay: number, value: any) => {
          // greets stackoverflow, i think
          return new Promise(resolve => setTimeout(resolve, delay, value));
        };
        await later(200, null); // courtesy delay
    

        // const result = await SearchQuery(scraperQuery)
        const result = await client.search(scraperQuery, 100, nextToken);
    
        const tweets: ITweet[] = [];
        if (result?.tweets) {
          for (const tweetId in result.tweets) {
            const tweet = result.tweets[tweetId];
            console.log(`  [🐦] ${tweet.id_str} ${tweet.full_text.substring(0, 70)}`);
            tweets.push({ id: tweet.id_str, full_text: tweet.full_text }); // Add other properties as needed
          }
        }
    
        // get the next page if there is one
        if (result?.nextToken) {
          const additionalTweets = await this.search(
            client,
            scraperQuery,
            result.nextToken,
            startDate,
            endDate
          );
          tweets.push(...additionalTweets);
        }
    
        return tweets;
      }

    /**
     * 
     * query Twitter API for all of an account's tweets
     * 
     * @warning There is a potential future problem with the /2/users/:id/tweets endpoint 
     *          in that pagination only returns the 3200 most recent tweets. 
     *          A potential workaround could be date-based search,
     *          but this is a future problem as we're only seeing ~2300 tweets as of May 1 2022
     * 
     *          see https://developer.twitter.com/en/docs/twitter-api/tweets/timelines/api-reference/get-users-id-tweets
     * 
     */
    async query(client, startDate, endDate): Promise<ITweet[]> {
        const scraperQuery = { terms: `from:${this.name} -filter:retweets filter:links since:${formatISO(startDate)} until:${formatISO(endDate)}` };
        const results = await this.search(client, scraperQuery, undefined, startDate, endDate)
        return results
    }

    async getAllTweetsOld(): Promise<ITweet[]> {
        const client = new TwitterClient['default']();
        await client.connect()
    
        let tweets = []
    
        // get all of Mel's tweets
        // because there is a limit of 3200 tweets
        // using regular pagination method, we instead
        // scrape twitter, making several searches using 5 day ranges
        // between the date that melody started streaming on CB and now
        for await (let [start, end] of this.getDateRange(this.startDate)) {
            const queryResults = await this.query(client, start, end);
            tweets.concat(queryResults)
        }
        return tweets
    }

    
    /**
     * get dates spaced a few days apart
     */
    async *getDateRange(startDate: Date): AsyncIterable<[Date, Date]> {
        let dateCounter = startDate;
        while (isBefore(dateCounter, new Date())) {
          const start = dateCounter;
          const end = add(dateCounter, { days: 8 });
          dateCounter = end;
          yield [start, end];
        }
    }
    
}