import puppeteer from 'puppeteer';
import { getRandomRoomFromCbHtml } from './cb.js';
import { z } from 'zod';
import { iPushServiceAuthSchema } from './headlessSchema.js';
import qs from 'qs';


export interface IChannelMap {
    [topic: string]: string;
}
  

export interface IPushServiceAuth {
    token: string;
    channels: IChannelMap;
    failures: Record<string, string>;
    token_request: {
        keyName: string;
        clientId: string;
        ttl: number;
        nonce: string;
        capability: string;
        timestamp: number;
        mac: string;
    }
    client_id: string;
    settings: {
        backend: string;
        flags: {
            pm_enabled: boolean;
            wowza_disabled: boolean;
            userlist_enabled: boolean;
            verify_enabled: boolean;
            fallback_eligible: boolean;
            is_live: boolean;
        }
        rest_host: string;
        realtime_host: string;
        fallback_hosts: string[];
    }
}

const roomSchema = z.string().min(3);

export function getSemverNumber(str: string): string {

    // Regular expression for matching SemVer numbers
    const semverRegex = /(?:\D|^)(\d+\.\d+\.\d+)(?=\D|$)/;

    // Use the regex to extract the SemVer number
    const match = str.match(semverRegex);

    // Check if there's a match and extract the SemVer number
    const semverNumber = match ? match[1] : null;
    return semverNumber;

}

export async function getRandomRoom(): Promise<string> {
    const browser = await puppeteer.launch({
        headless: 'new'
    });
    const page = await browser.newPage();
    await page.goto('https://chaturbate.com')
    const html = await page.content();
    await browser.close();
    const room = await getRandomRoomFromCbHtml(html);
    roomSchema.parse(room);
    return room;
}

export async function getAblyAgent(): Promise<string> {
    const browser = await puppeteer.launch({
        headless: 'new'
    });
    const page = await browser.newPage();
    await page.goto('https://chaturbate.com')
    const html = await page.content();
    const room = await getRandomRoomFromCbHtml(html);
    roomSchema.parse(room);
    await page.setRequestInterception(true);

    const url: string = await new Promise((resolve) => {
        page.on('request', interceptRequest => {
            const url = interceptRequest.url();
            if (url.startsWith('https://realtime')) {
                console.log(`request! ${interceptRequest.url()} ${JSON.stringify(interceptRequest.headers(), null, 2)}`);
                resolve(interceptRequest.url());
            }
            interceptRequest.continue();
        });

        page.goto(`https://chaturbate.com/${room}`);
    })

    const query = qs.parse(url);
    console.log(query);
    await browser.close();
    return decodeURIComponent(query?.agent);
}

export async function getAblyAgentVersionSemver(): Promise<string> {
    const agentString = await getAblyAgent();
    return getSemverNumber(agentString);
}


export async function getPushServiceAuth(roomUrl: string): Promise<IPushServiceAuth> {
    if (!roomUrl) throw new Error('roomUrl is a required param, but it was undefined.');

    const roomUrlUrl = new URL(roomUrl);
    const browser = await puppeteer.launch({
        headless: 'new'
    });
    const page = await browser.newPage();


    await page.setRequestInterception(true);
    page.on('request', interceptRequest => {
        // const url = interceptRequest.url();
        // if (url === 'https://chaturbate.com/push_service/auth/') {
        //     console.log(`request! ${interceptRequest.url()} ${JSON.stringify(interceptRequest.headers(), null, 2)}`);
        // }
        interceptRequest.continue();
    })

    const json = await new Promise(async (resolve) => {
        page.on('response', async (interceptResponse) => {
            const url = interceptResponse.url();
            if (url === 'https://chaturbate.com/push_service/auth/') {
                // console.log(`response! ${url}`);
                const json = await interceptResponse.json();
                resolve(json);
            }
        })
        console.log(`>> navigating to ${roomUrlUrl.toString()}`)
        try {
            await page.goto(roomUrlUrl.toString());
        } catch (e) {
            //  It's likely not an issue because we might already have the data we need.
            if (!e.message.includes('Navigation failed because browser has disconnected')) {
                console.log(`>> caught an error during navigation.`);
                console.log(e);
            }
        }
    })

    let psa;
    try {
        psa = iPushServiceAuthSchema.parse(json);
        console.log(`>> got psa`);
    } catch (e) {
        console.log(`>> error!!!! ${e}`);
        if (e instanceof Error) {
            console.error(e)
        }
    } finally {
        console.log('>> closing the browser')
        await browser.close()
    }

    return psa;

}